
#include <avr/io.h>	
#include <avr/interrupt.h> 
#include <util/delay.h>
#include <stdlib.h>
#include <inttypes.h>
#include <stdbool.h>
#include <stdio.h>

#include <lcd.h>

/*
 * Port usage
 *
 *
 * PB0 IN  RANGE BTN
 * PB1 IN  WAVEFORM BTN
 * PB2 
 * PB3 --- AVR ISP MOSI
 * PB4 --- AVR ISP MISO
 * PB5 --- AVR ISP CLK
 * PB6 XTAL
 * PB7 XTAL
 *
 * PC0 OUT RANGE0
 * PC1 OUT RANGE1
 * PC2 OUT WF SINE ENABLE
 * PC3 OUT WF TRIANGLE ENABLE
 * PC4 OUT WF SQUARE ENABLE
 * PC5 OUT WF SINE SHAPE
 * PC6 RESET
 *
 * PD0 OUT LCD DB4
 * PD1 OUT LCD DB5
 * PD2 OUT LCD DB6
 * PD3 OUT LCD DB7
 * PD4 OUT LCD RS
 * PD5 IN  T1
 * PD6 OUT LCD EN
 * PD7 
 *
 */

volatile static uint8_t t1_tick = 0; 
volatile static uint8_t t1_ovf_cnt = 0;

#define BTN_INERTIA 25500

ISR (TIMER1_OVF_vect)
{
	t1_ovf_cnt++;
}

ISR(TIMER2_COMPA_vect) 
{
	t1_tick++;
}

static void init (void) {

	// port directions

	DDRB = 0;
	DDRC = (1<<4) | (1<<3) | (1<<2) | (1<<1) | (1<<0) ;
	DDRD = (1<<6) | (1<<4) | (1<<3) | (1<<2) | (1<<1) | (1<<0);

	PORTB = 0;
	PORTC = 0;
	PORTD = 0;

	// lcd

	lcd_init();

	// timer 1 (T1 clk source)

	TIMSK1 |= (1<<TOIE1); // overflow interrupt enable
	TCCR1B |= (1<<CS10) | (1<<CS12) | (1<<CS11); // External clock source on T1 pin. Clock on rising edge.
	//TIFR1 = (1<<TOIE1) + (1<<ICIE1);    // Schon aktive Interrupts löschen, Mega32: TIFR

	// timer 2
	// f = 20Mhz / 1024 / 125 = 156.25 Hz => T = 6.4ms

	TCCR2A |= (1<<WGM21); // CTC Mode Aktivierung Timer2
	TCCR2A |= (1<<COM2A0) | (1<<COM2A1); // set OCR Flag bei "Compare Match"
	TCCR2B |= (1<<CS22) | (1<<CS21) | (1<<CS20); //  Prescaler Timer2 auf 1024 setzen
	OCR2A  |= 124; // Output Compare register Timer2 auf 124 (für Loop 0-124 => 125)
	TIMSK2 |= (1<<OCIE2A); // Enable comp match flag interrupt Timer 2 

	sei();

}

int main(void) {

	static char anz[20];
	static uint16_t btn_inertia = 0;
	static uint8_t range = 0;
	static uint8_t wf = 0;
	static char key_down = 0;
	static uint32_t freq = 0;

	init();

	lcd_setcursor( 0, 1 );
	lcd_string("SINE  10-100 kHz");
	lcd_setcursor( 0, 2 );
	lcd_string("                ");

	// preselect sine waveform
	PORTC |= (1 << 2);
	PORTC |= (0 << 3);
	PORTC |= (0 << 4);
	PORTC |= (1 << 5);

	// preselect range
	PORTC |= (1 << 1);

	while(1) {


		if (btn_inertia) {
			btn_inertia--; // keybounce
		} else {

			if (key_down) {
				if ((PINB & 3) == 0) {
					key_down = 0;
					btn_inertia = BTN_INERTIA;
				}
			} else {

				if (PINB & 1) { // range button
					btn_inertia = BTN_INERTIA;
					key_down = 1;

					range = (range + 1) % 4;

					PORTC &= ~3;
					PORTC |= range;

					lcd_setcursor( 5, 1 );
					switch (range) {
						case 0:
							lcd_string ("  0 - 1 kHz");
							break;
						case 1:
							lcd_string (" 1 - 10 kHz");
							break;
						case 2:
							lcd_string (" 10-100 kHz");
							break;
						case 3:
							lcd_string ("0.1-1.0 MHz");
							break;
					}

				} else if (PINB & 2) { // waveform button
					btn_inertia = BTN_INERTIA;
					key_down = 1;

					wf += 1;
					if (wf > 3) {
						wf = 0;
					}

					PORTC &= ~( (1<<5) | (1<<4) | (1<<3) | (1<<2) );

					lcd_setcursor( 0, 1 );
					switch (wf) {
						case 0:
							lcd_string ("SINE");
							PORTC |= (1 << 2);
							PORTC |= (0 << 3);
							PORTC |= (0 << 4);
							PORTC |= (1 << 5);
							break;
						case 1:
							lcd_string ("SQRE");
							PORTC |= (0 << 2);
							PORTC |= (0 << 3);
							PORTC |= (1 << 4);
							PORTC |= (0 << 5);
							break;
						case 2:
							lcd_string ("TRIA");
							PORTC |= (0 << 2);
							PORTC |= (1 << 3);
							PORTC |= (0 << 4);
							PORTC |= (0 << 5);
							break;
						case 3:
							lcd_string ("CAP ");
							PORTC |= (0 << 2);
							PORTC |= (1 << 3);
							PORTC |= (0 << 4);
							PORTC |= (0 << 5);
							break;
					}
				}
			}
		}

		if (t1_tick==125) { // once a second

			uint32_t count = TCNT1;
			uint32_t z = t1_ovf_cnt ;

			t1_tick=0;
			t1_ovf_cnt = 0;
			TCNT1 = 0;

			if (wf != 3) {
				freq = count + 65536l * z;
				ltoa(freq, anz, 10);
				lcd_setcursor( 0, 2 );
				lcd_string("                ");
				lcd_setcursor( 0, 2 );
				lcd_string( anz );
				lcd_setcursor( 8, 2 );
				lcd_string( "Hz" );
			} else {

				// measure capacitance

				uint32_t f2 = count + 65536l * z;

				uint32_t cx = (freq * 1000l / f2) - 1000;

				ltoa(cx, anz, 10);
				lcd_setcursor( 0, 2 );
				lcd_string("                ");
				lcd_setcursor( 0, 2 );
				lcd_string( anz );
				lcd_setcursor( 8, 2 );
				lcd_string( "pF" );
			}
		}
	}

}		

